@@ -29,17 +29,13 @@
 
 #define MAX_PLAYER_ICONS 8
 
-#ifdef CATEGORY_FILE
-#define CATEGORY_FACTORY_FUNC columns_factory_category
-#else
-#define CATEGORY_FACTORY_FUNC NULL
-#endif
-
 typedef GtkTreeViewColumn * (*FactoryFunc) (GvaGameStoreColumn);
+typedef gboolean (*TooltipFunc) (GtkTreeModel *, GtkTreeIter *, GtkTooltip *);
 
 static GdkPixbuf *
 columns_get_icon_name (const gchar *icon_name)
 {
+        static GHashTable *cache = NULL;
         GtkIconTheme *icon_theme;
         GdkPixbuf *pixbuf;
         GdkPixbuf *scaled;
@@ -47,6 +43,16 @@ columns_get_icon_name (const gchar *icon_name)
         gint size;
         GError *error = NULL;
 
+        if (G_UNLIKELY (cache == NULL))
+                cache = g_hash_table_new_full (
+                        g_str_hash, g_str_equal,
+                        (GDestroyNotify) g_free,
+                        (GDestroyNotify) g_object_unref);
+
+        scaled = g_hash_table_lookup (cache, icon_name);
+        if (scaled != NULL)
+                return scaled;
+
         icon_theme = gtk_icon_theme_get_default ();
         valid = gtk_icon_size_lookup (GTK_ICON_SIZE_MENU, &size, NULL);
         g_assert (valid);
@@ -63,6 +69,8 @@ columns_get_icon_name (const gchar *icon_name)
         scaled = gdk_pixbuf_scale_simple (
                 pixbuf, size, size, GDK_INTERP_BILINEAR);
 
+        g_hash_table_insert (cache, g_strdup (icon_name), scaled);
+
         g_object_unref (pixbuf);
 
         return scaled;
@@ -79,11 +87,7 @@ columns_favorite_clicked_cb (GvaCellRendererPixbuf *renderer,
         /* The row that was clicked is not yet selected.  We need to
          * select it first so that gva_tree_view_get_selected_game()
          * returns the correct name. */
-#if GTK_CHECK_VERSION (2, 12, 0)
         widget = gtk_tree_view_column_get_tree_view (column);
-#else
-        widget = column->tree_view;
-#endif
         gtk_tree_view_set_cursor (GTK_TREE_VIEW (widget), path, NULL, FALSE);
         gtk_widget_grab_focus (widget);
 
@@ -109,11 +113,7 @@ columns_comment_edited_cb (GtkCellRendererText *renderer,
         gint column_id;
         gboolean valid;
 
-#if GTK_CHECK_VERSION (2, 12, 0)
         widget = gtk_tree_view_column_get_tree_view (column);
-#else
-        widget = column->tree_view;
-#endif
         column_id = gtk_tree_view_column_get_sort_column_id (column);
         model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));
         path = gtk_tree_path_new_from_string (path_string);
@@ -125,40 +125,73 @@ columns_comment_edited_cb (GtkCellRendererText *renderer,
                 GTK_TREE_STORE (model), &iter, column_id, new_text, -1);
 }
 
+static void
+columns_bios_set_properties (GtkTreeViewColumn *column,
+                             GtkCellRenderer *renderer,
+                             GtkTreeModel *model,
+                             GtkTreeIter *iter)
+{
+        GvaGameStoreColumn column_id;
+        gchar *bios, *cp;
+        gsize length;
+
+        column_id = gtk_tree_view_column_get_sort_column_id (column);
+        gtk_tree_model_get (model, iter, column_id, &bios, -1);
+
+        /* A lot of BIOS descriptions end with the word "BIOS".
+         * Strip it off, since the column is already titled "BIOS". */
+        length = strlen (bios);
+        cp = bios + length - 5;
+        if (g_ascii_strcasecmp (cp, " BIOS") == 0)
+                *cp = '\0';
+
+        g_object_set (renderer, "text", bios, NULL);
+
+        g_free (bios);
+}
+
 static void
 columns_driver_status_set_properties (GtkTreeViewColumn *column,
                                       GtkCellRenderer *renderer,
                                       GtkTreeModel *model,
                                       GtkTreeIter *iter)
 {
-        GvaGameStoreColumn column_id;
         GdkPixbuf *pixbuf = NULL;
-        gchar *status;
+        gchar *driver_status;
+        gchar *driver_emulation;
+        gchar *driver_protection;
 
-        column_id = gtk_tree_view_column_get_sort_column_id (column);
-        gtk_tree_model_get (model, iter, column_id, &status, -1);
+        gtk_tree_model_get (
+                model, iter,
+                GVA_GAME_STORE_COLUMN_DRIVER_STATUS, &driver_status,
+                GVA_GAME_STORE_COLUMN_DRIVER_EMULATION, &driver_emulation,
+                GVA_GAME_STORE_COLUMN_DRIVER_PROTECTION, &driver_protection,
+                -1);
 
         /* XXX For the "good" icon I want some kind of positive symbol; a
-         *     checkmark or thumbs up or something.  The "emblem-default"
-         *     image in gnome-icon-theme seems to be the closest match at
-         *     this time (a green checkmark), but may not be suitable for
-         *     all icon themes. */
+         *     checkmark or thumbs up or something.  The GTK_STOCK_APPLY 
+         *     image seems to be the closest match at this time (a green
+         *     checkmark), but may not be suitable for all icon themes. */
 
-        if (strcmp (status, "good") == 0)
-                pixbuf = columns_get_icon_name ("emblem-default");
-        else if (strcmp (status, "imperfect") == 0)
-                pixbuf = columns_get_icon_name ("dialog-warning");
-        else if (strcmp (status, "preliminary") == 0)
-                pixbuf = columns_get_icon_name ("dialog-error");
+        /* Remember, we don't own the pixbuf reference. */
+        if (strcmp (driver_status, "good") == 0)
+                pixbuf = columns_get_icon_name (GTK_STOCK_APPLY);
+        else if (strcmp (driver_emulation, "preliminary") == 0)
+                pixbuf = columns_get_icon_name (GTK_STOCK_DIALOG_ERROR);
+        else if (strcmp (driver_protection, "preliminary") == 0)
+                pixbuf = columns_get_icon_name (GTK_STOCK_DIALOG_ERROR);
+        else if (strcmp (driver_status, "imperfect") == 0)
+                pixbuf = columns_get_icon_name (GTK_STOCK_DIALOG_WARNING);
+        else if (strcmp (driver_status, "preliminary") == 0)
+                pixbuf = columns_get_icon_name (GTK_STOCK_DIALOG_WARNING);
 
         g_object_set (
                 renderer, "pixbuf", pixbuf,
                 "visible", (pixbuf != NULL), NULL);
 
-        if (pixbuf != NULL)
-                g_object_unref (pixbuf);
-
-        g_free (status);
+        g_free (driver_status);
+        g_free (driver_emulation);
+        g_free (driver_protection);
 }
 
 static void
@@ -225,8 +258,12 @@ columns_time_set_properties (GtkTreeViewColumn *column,
         g_value_unset (&value);
 }
 
+/*****************************************************************************
+ * Column Factory Callbacks
+ *****************************************************************************/
+
 static GtkTreeViewColumn *
-columns_factory_category (GvaGameStoreColumn column_id)
+columns_factory_bios (GvaGameStoreColumn column_id)
 {
         GtkTreeViewColumn *column;
         GtkCellRenderer *renderer;
@@ -234,7 +271,27 @@ columns_factory_category (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Category"));
+
+        renderer = gtk_cell_renderer_text_new ();
+        gtk_tree_view_column_pack_start (column, renderer, TRUE);
+
+        gtk_tree_view_column_set_cell_data_func (
+                column, renderer, (GtkTreeCellDataFunc)
+                columns_bios_set_properties, NULL, NULL);
+
+        return column;
+}
+
+static GtkTreeViewColumn *
+columns_factory_category (GvaGameStoreColumn column_id)
+{
+#ifdef CATEGORY_FILE
+        GtkTreeViewColumn *column;
+        GtkCellRenderer *renderer;
+
+        column = gtk_tree_view_column_new ();
+        gtk_tree_view_column_set_reorderable (column, TRUE);
+        gtk_tree_view_column_set_sort_column_id (column, column_id);
 
         renderer = gtk_cell_renderer_text_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -243,6 +300,9 @@ columns_factory_category (GvaGameStoreColumn column_id)
                 column, renderer, "text", column_id);
 
         return column;
+#else
+        return NULL;
+#endif
 }
 
 static GtkTreeViewColumn *
@@ -255,7 +315,6 @@ columns_factory_comment (GvaGameStoreColumn column_id)
         gtk_tree_view_column_set_expand (column, TRUE);
         gtk_tree_view_column_set_reorderable (column, FALSE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Comment"));
 
         renderer = gtk_cell_renderer_text_new ();
         g_object_set (renderer, "editable", TRUE, NULL);
@@ -282,7 +341,6 @@ columns_factory_description (GvaGameStoreColumn column_id)
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
         gtk_tree_view_column_set_spacing (column, 3);
-        gtk_tree_view_column_set_title (column, _("Title"));
 
         renderer = gtk_cell_renderer_text_new ();
         g_object_set (renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
@@ -303,7 +361,6 @@ columns_factory_driver_status (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Status"));
 
         renderer = gtk_cell_renderer_pixbuf_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -322,12 +379,12 @@ columns_factory_favorite (GvaGameStoreColumn column_id)
         GtkCellRenderer *renderer;
         GdkPixbuf *pixbuf;
 
+        /* Remember, we don't own the pixbuf reference. */
         pixbuf = columns_get_icon_name ("emblem-favorite");
 
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Favorite"));
 
         renderer = gva_cell_renderer_pixbuf_new ();
         g_object_set (renderer, "pixbuf", pixbuf, NULL);
@@ -340,9 +397,6 @@ columns_factory_favorite (GvaGameStoreColumn column_id)
                 renderer, "clicked",
                 G_CALLBACK (columns_favorite_clicked_cb), column);
 
-        if (pixbuf != NULL)
-                g_object_unref (pixbuf);
-
         return column;
 }
 
@@ -353,12 +407,12 @@ columns_factory_input_players (GvaGameStoreColumn column_id)
         GdkPixbuf *pixbuf;
         gint ii;
 
+        /* Remember, we don't own the pixbuf reference. */
         pixbuf = columns_get_icon_name ("stock_person");
 
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Players"));
 
         for (ii = 0; ii < MAX_PLAYER_ICONS; ii++)
         {
@@ -374,9 +428,6 @@ columns_factory_input_players (GvaGameStoreColumn column_id)
                         GINT_TO_POINTER (ii), NULL);
         }
 
-        if (pixbuf != NULL)
-                g_object_unref (pixbuf);
-
         return column;
 }
 
@@ -389,7 +440,6 @@ columns_factory_manufacturer (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Manufacturer"));
 
         renderer = gtk_cell_renderer_text_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -409,7 +459,6 @@ columns_factory_name (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("ROM Name"));
 
         renderer = gtk_cell_renderer_text_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -427,12 +476,12 @@ columns_factory_sampleset (GvaGameStoreColumn column_id)
         GtkCellRenderer *renderer;
         GdkPixbuf *pixbuf;
 
+        /* Remember, we don't own the pixbuf reference. */
         pixbuf = columns_get_icon_name ("emblem-sound");
 
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Samples"));
 
         renderer = gtk_cell_renderer_pixbuf_new ();
         g_object_set (renderer, "pixbuf", pixbuf, NULL);
@@ -442,9 +491,6 @@ columns_factory_sampleset (GvaGameStoreColumn column_id)
                 column, renderer, (GtkTreeCellDataFunc)
                 columns_sampleset_set_properties, NULL, NULL);
 
-        if (pixbuf != NULL)
-                g_object_unref (pixbuf);
-
         return column;
 }
 
@@ -457,7 +503,6 @@ columns_factory_sourcefile (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Driver"));
 
         renderer = gtk_cell_renderer_text_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -477,7 +522,6 @@ columns_factory_time (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, FALSE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Played On"));
 
         renderer = gtk_cell_renderer_text_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -498,7 +542,6 @@ columns_factory_year (GvaGameStoreColumn column_id)
         column = gtk_tree_view_column_new ();
         gtk_tree_view_column_set_reorderable (column, TRUE);
         gtk_tree_view_column_set_sort_column_id (column, column_id);
-        gtk_tree_view_column_set_title (column, _("Year"));
 
         renderer = gtk_cell_renderer_text_new ();
         gtk_tree_view_column_pack_start (column, renderer, TRUE);
@@ -509,33 +552,327 @@ columns_factory_year (GvaGameStoreColumn column_id)
         return column;
 }
 
+/*****************************************************************************
+ * Column Tooltip Callbacks
+ *****************************************************************************/
+
+static gboolean
+columns_tooltip_driver_status (GtkTreeModel *model,
+                               GtkTreeIter *iter,
+                               GtkTooltip *tooltip)
+{
+        GtkWidget *table;
+        GtkWidget *widget;
+        const gchar *text;
+        const gchar *stock_id;
+        gchar *driver_status;
+        gchar *driver_emulation;
+        gchar *driver_color;
+        gchar *driver_sound;
+        gchar *driver_graphic;
+        gchar *driver_cocktail;
+        gchar *driver_protection;
+        gboolean show_tooltip;
+
+        gtk_tree_model_get (
+                model, iter,
+                GVA_GAME_STORE_COLUMN_DRIVER_STATUS, &driver_status,
+                GVA_GAME_STORE_COLUMN_DRIVER_EMULATION, &driver_emulation,
+                GVA_GAME_STORE_COLUMN_DRIVER_COLOR, &driver_color,
+                GVA_GAME_STORE_COLUMN_DRIVER_SOUND, &driver_sound,
+                GVA_GAME_STORE_COLUMN_DRIVER_GRAPHIC, &driver_graphic,
+                GVA_GAME_STORE_COLUMN_DRIVER_COCKTAIL, &driver_cocktail,
+                GVA_GAME_STORE_COLUMN_DRIVER_PROTECTION, &driver_protection,
+                -1);
+
+        show_tooltip =
+                strcmp (driver_status, "imperfect") == 0 ||
+                strcmp (driver_status, "preliminary") == 0;
+        if (!show_tooltip)
+                goto exit;
+
+        if (strcmp (driver_emulation, "preliminary") == 0)
+                stock_id = GTK_STOCK_DIALOG_ERROR;
+        else if (strcmp (driver_protection, "preliminary") == 0)
+                stock_id = GTK_STOCK_DIALOG_ERROR;
+        else
+                stock_id = GTK_STOCK_DIALOG_WARNING;
+
+        table = gtk_table_new (9, 2, FALSE);
+        gtk_table_set_col_spacings (GTK_TABLE (table), 6);
+        gtk_table_set_row_spacing (GTK_TABLE (table), 0, 6);
+        gtk_widget_show (table);
+
+        widget = gtk_image_new_from_stock (stock_id, GTK_ICON_SIZE_DND);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.5, 0.0);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 0, 1, 0, 9,
+                0, GTK_EXPAND | GTK_FILL, 0, 0);
+        gtk_widget_show (widget);
+
+        /* The same text is in gnome-video-arcade.glade,
+         * so it has to be translated with markup anyway. */
+        text = _("<b>There are known problems with this game:</b>");
+        widget = gtk_label_new (text);
+        gtk_label_set_use_markup (GTK_LABEL (widget), TRUE);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 0, 1,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        gtk_widget_show (widget);
+
+        /* The labels begin with a UTF-8 encoded bullet character. */
+
+        text = _("• The colors aren't 100% accurate.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 1, 2,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_color, "imperfect") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• The colors are completely wrong.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 2, 3,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_color, "preliminary") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• The video emulation isn't 100% accurate.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 3, 4,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_graphic, "imperfect") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• The sound emulation isn't 100% accurate.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 4, 5,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_sound, "imperfect") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• The game lacks sound.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 5, 6,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_sound, "preliminary") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• Screen flipping in cocktail mode is not supported.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 6, 7,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_cocktail, "preliminary") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• <b>THIS GAME DOESN'T WORK.</b>");
+        widget = gtk_label_new (text);
+        gtk_label_set_use_markup (GTK_LABEL (widget), TRUE);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 7, 8,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_emulation, "preliminary") == 0)
+                gtk_widget_show (widget);
+
+        text = _("• The game has protection which isn't fully emulated.");
+        widget = gtk_label_new (text);
+        gtk_misc_set_alignment (GTK_MISC (widget), 0.0, 0.5);
+        gtk_table_attach (
+                GTK_TABLE (table), widget, 1, 2, 8, 9,
+                GTK_EXPAND | GTK_FILL, 0, 0, 0);
+        if (strcmp (driver_protection, "preliminary") == 0)
+                gtk_widget_show (widget);
+
+        gtk_tooltip_set_custom (tooltip, table);
+
+exit:
+        g_free (driver_status);
+        g_free (driver_emulation);
+        g_free (driver_color);
+        g_free (driver_sound);
+        g_free (driver_graphic);
+        g_free (driver_cocktail);
+        g_free (driver_protection);
+
+        return show_tooltip;
+}
+
+static gboolean
+columns_tooltip_favorite (GtkTreeModel *model,
+                          GtkTreeIter *iter,
+                          GtkTooltip *tooltip)
+{
+        const gchar *text;
+        gboolean favorite;
+
+        gtk_tree_model_get (
+                model, iter, GVA_GAME_STORE_COLUMN_FAVORITE, &favorite, -1);
+
+        if (favorite)
+                text = _("Click here to remove from favorites");
+        else
+                text = _("Click here to add to favorites");
+
+        gtk_tooltip_set_text (tooltip, text);
+
+        return TRUE;
+}
+
+static gboolean
+columns_tooltip_input_players (GtkTreeModel *model,
+                               GtkTreeIter *iter,
+                               GtkTooltip *tooltip)
+{
+        const gchar *text;
+        gint max_players;
+
+        gtk_tree_model_get (
+                model, iter,
+                GVA_GAME_STORE_COLUMN_INPUT_PLAYERS, &max_players, -1);
+
+        /* Keep this in sync with MAX_PLAYER_ICONS. */
+        switch (max_players)
+        {
+                case 1:
+                        text = _("One player only");
+                        break;
+
+                case 2:
+                        text = _("One or two players");
+                        break;
+
+                case 3:
+                        text = _("Up to three players");
+                        break;
+
+                case 4:
+                        text = _("Up to four players");
+                        break;
+
+                case 5:
+                        text = _("Up to five players");
+                        break;
+
+                case 6:
+                        text = _("Up to six players");
+                        break;
+
+                case 7:
+                        text = _("Up to seven players");
+                        break;
+
+                case 8:
+                        text = _("Up to eight players");
+                        break;
+
+                default:
+                        return FALSE;
+        }
+
+        gtk_tooltip_set_text (tooltip, text);
+
+        return TRUE;
+}
+
+static gboolean
+columns_tooltip_summary (GtkTreeModel *model,
+                         GtkTreeIter *iter,
+                         GtkTooltip *tooltip)
+{
+        gchar *description;
+        gchar *manufacturer;
+        gchar *year;
+        gchar *markup;
+
+        gtk_tree_model_get (
+                model, iter,
+                GVA_GAME_STORE_COLUMN_DESCRIPTION, &description,
+                GVA_GAME_STORE_COLUMN_MANUFACTURER, &manufacturer,
+                GVA_GAME_STORE_COLUMN_YEAR, &year, -1);
+
+        if (description == NULL || *description == '\0')
+                description = g_strdup (_("(Game Description Unknown)"));
+
+        if (manufacturer == NULL || *manufacturer == '\0')
+                manufacturer = g_strdup (_("(Manufacturer Unknown)"));
+
+        if (year == NULL || *year == '\0')
+                year = g_strdup (_("(Year Unknown)"));
+
+        markup = g_markup_printf_escaped (
+                "<b>%s</b>\n<small>%s %s</small>",
+                description, year, manufacturer);
+        gtk_tooltip_set_markup (tooltip, markup);
+        g_free (markup);
+
+        g_free (description);
+        g_free (manufacturer);
+        g_free (year);
+
+        return TRUE;
+}
+
 static struct
 {
         const gchar *name;
+        const gchar *title;
         FactoryFunc factory;
+        TooltipFunc tooltip;
 }
 column_info[GVA_GAME_STORE_NUM_COLUMNS] =
 {
-        { "name",               columns_factory_name },
-        { "category",           CATEGORY_FACTORY_FUNC },
-        { "favorite",           columns_factory_favorite },
-        { "sourcefile",         columns_factory_sourcefile },
+        { "name",               N_("ROM Name"),
+                                columns_factory_name },
+        { "bios",               N_("BIOS"),
+                                columns_factory_bios },
+        { "category",           N_("Category"),
+                                columns_factory_category },
+        { "favorite",           N_("Favorite"),
+                                columns_factory_favorite,
+                                columns_tooltip_favorite },
+        { "sourcefile",         N_("Driver"),
+                                columns_factory_sourcefile },
+        { "isbios",             NULL },
         { "runnable",           NULL },
         { "cloneof",            NULL },
         { "romof",              NULL },
         { "romset",             NULL },
         { "sampleof",           NULL },
-        { "sampleset",          columns_factory_sampleset },
-        { "description",        columns_factory_description },
-        { "year",               columns_factory_year },
-        { "manufacturer",       columns_factory_manufacturer },
+        { "sampleset",          N_("Samples"),
+                                columns_factory_sampleset },
+        { "description",        N_("Title"),
+                                columns_factory_description,
+                                columns_tooltip_summary },
+        { "year",               N_("Year"),
+                                columns_factory_year,
+                                columns_tooltip_summary },
+        { "manufacturer",       N_("Manufacturer"),
+                                columns_factory_manufacturer,
+                                columns_tooltip_summary },
         { "sound_channels",     NULL },
         { "input_service",      NULL },
         { "input_tilt",         NULL },
-        { "input_players",      columns_factory_input_players },
+        { "input_players",      N_("Players"),
+                                columns_factory_input_players,
+                                columns_tooltip_input_players },
         { "input_buttons",      NULL },
         { "input_coins",        NULL },
-        { "driver_status",      columns_factory_driver_status },
+        { "driver_status",      N_("Status"),
+                                columns_factory_driver_status,
+                                columns_tooltip_driver_status },
         { "driver_emulation",   NULL },
         { "driver_color",       NULL },
         { "driver_sound",       NULL },
@@ -544,10 +881,12 @@ column_info[GVA_GAME_STORE_NUM_COLUMNS] =
         { "driver_protection",  NULL },
         { "driver_savestate",   NULL },
         { "driver_palettesize", NULL },
-        { "comment",            columns_factory_comment },
+        { "comment",            N_("Comment"),
+                                columns_factory_comment },
         { "inode",              NULL },
         { "inpfile",            NULL },
-        { "time",               columns_factory_time }
+        { "time",               N_("Played On"),
+                                columns_factory_time }
 };
 
 static gchar *default_column_order[] =
@@ -559,6 +898,7 @@ static gchar *default_column_order[] =
 #ifdef CATEGORY_FILE
         "category",
 #endif
+        "bios",
         "driver_status",
         "input_players",
         "name",
@@ -579,6 +919,7 @@ GtkTreeViewColumn *
 gva_columns_new_from_id (GvaGameStoreColumn column_id)
 {
         GtkTreeViewColumn *column;
+        const gchar *title;
 
         g_return_val_if_fail (column_id >= 0, NULL);
         g_return_val_if_fail (column_id < G_N_ELEMENTS (column_info), NULL);
@@ -586,9 +927,13 @@ gva_columns_new_from_id (GvaGameStoreColumn column_id)
 
         column = column_info[column_id].factory (column_id);
 
-        g_object_set_data (
-                G_OBJECT (column), "name",
-                (gpointer) column_info[column_id].name);
+        title = gettext (column_info[column_id].title);
+        gtk_tree_view_column_set_title (column, title);
+
+        if (column != NULL)
+                g_object_set_data (
+                        G_OBJECT (column), "name",
+                        (gpointer) column_info[column_id].name);
 
         return column;
 }
@@ -661,6 +1006,23 @@ gva_columns_lookup_name (GvaGameStoreColumn column_id)
         return column_info[column_id].name;
 }
 
+/**
+ * gva_columns_lookup_title:
+ * @column_id: the ID of the column to lookup
+ *
+ * Looks up the column title corresponding to @column_id.
+ *
+ * Returns: the column title, or @NULL if not found
+ **/
+const gchar *
+gva_columns_lookup_title (GvaGameStoreColumn column_id)
+{
+        if (CLAMP (column_id, 0, G_N_ELEMENTS (column_info)) != column_id)
+                return NULL;
+
+        return column_info[column_id].title;
+}
+
 /* Helper for gva_columns_load() */
 static gboolean
 columns_load_remove_name (GSList **p_list, const gchar *name)
@@ -926,13 +1288,81 @@ gva_columns_get_names_full (GtkTreeView *view)
 
         for (iter = names; iter != NULL; iter = iter->next)
         {
-                /* XXX Nothing to do yet.
                 const gchar *column_name = iter->data;
 
+                /* Need to know if a ROM set is not a game but a BIOS,
+                 * so we can exclude it from the game list. */
+                if (strcmp (column_name, "name") == 0)
+                        columns_add_dependency (&names, "isbios");
+
+                /* Any of the description, manufacturer, and year columns
+                 * require the other two for the summary tooltip. */
                 if (strcmp (column_name, "description") == 0)
-                        columns_add_dependency (&names, "driver_status");
-                */
+                {
+                        columns_add_dependency (&names, "manufacturer");
+                        columns_add_dependency (&names, "year");
+                }
+                else if (strcmp (column_name, "manufacturer") == 0)
+                {
+                        columns_add_dependency (&names, "description");
+                        columns_add_dependency (&names, "year");
+                }
+                else if (strcmp (column_name, "year") == 0)
+                {
+                        columns_add_dependency (&names, "description");
+                        columns_add_dependency (&names, "manufacturer");
+                }
+
+                /* We give detailed information in the status tooltip. */
+                if (strcmp (column_name, "driver_status") == 0)
+                {
+                        columns_add_dependency (&names, "driver_emulation");
+                        columns_add_dependency (&names, "driver_color");
+                        columns_add_dependency (&names, "driver_sound");
+                        columns_add_dependency (&names, "driver_graphic");
+                        columns_add_dependency (&names, "driver_cocktail");
+                        columns_add_dependency (&names, "driver_protection");
+                }
         }
 
         return names;
 }
+
+/**
+ * gva_columns_query_tooltip:
+ * @column: a #GtkTreeViewColumn
+ * @path: a #GtkTreePath
+ * @tooltip: a #GtkTooltip
+ *
+ * Configures @tooltip for the given @column and @path.
+ *
+ * Returns: %TRUE if the tooltip should be shown
+ **/
+gboolean
+gva_columns_query_tooltip (GtkTreeViewColumn *column,
+                           GtkTreePath *path,
+                           GtkTooltip *tooltip)
+{
+        TooltipFunc tooltip_func;
+        GvaGameStoreColumn column_id;
+        GtkTreeModel *model;
+        GtkWidget *widget;
+        GtkTreeIter iter;
+        gboolean valid;
+
+        g_return_val_if_fail (GTK_IS_TREE_VIEW_COLUMN (column), FALSE);
+        g_return_val_if_fail (path != NULL, FALSE);
+        g_return_val_if_fail (GTK_IS_TOOLTIP (tooltip), FALSE);
+
+        column_id = gtk_tree_view_column_get_sort_column_id (column);
+        g_return_val_if_fail (column_id >= 0, FALSE);
+        g_return_val_if_fail (column_id < G_N_ELEMENTS (column_info), FALSE);
+
+        tooltip_func = column_info[column_id].tooltip;
+        widget = gtk_tree_view_column_get_tree_view (column);
+        model = gtk_tree_view_get_model (GTK_TREE_VIEW (widget));
+        valid = gtk_tree_model_get_iter (model, &iter, path);
+        g_return_val_if_fail (valid, FALSE);
+
+        return tooltip_func ? tooltip_func (model, &iter, tooltip) : FALSE;
+}
