@@ -1,4 +1,4 @@
-/* Copyright 2007 Matthew Barnes
+/* Copyright 2007, 2008 Matthew Barnes
  *
  * This file is part of GNOME Video Arcade.
  *
@@ -26,15 +26,76 @@
 #include "gva-favorites.h"
 #include "gva-game-store.h"
 #include "gva-main.h"
+#include "gva-mame.h"
+#include "gva-preferences.h"
 #include "gva-ui.h"
+#include "gva-util.h"
 
 #define SQL_SELECT_GAMES \
-        "SELECT %s FROM game WHERE (romset IN ('good', 'best available'))"
+        "SELECT %s FROM available"
+
+void
+tree_view_add_search_expression (GString *expression)
+{
+        gchar *column_name = NULL;
+        gchar *search_text = NULL;
+
+        if (gva_main_get_last_selected_match (&column_name, &search_text))
+                g_string_append_printf (
+                        expression, "%s == '%s'", column_name, search_text);
+        else
+        {
+                search_text = gva_main_get_last_search_text ();
+                if (search_text != NULL && *search_text != '\0')
+                        g_string_append_printf (
+                                expression,
+                                "(name LIKE '%s' OR "
+                                "bios MATCH '%s' OR "
+                                "category MATCH '%s' OR "
+                                "sourcefile LIKE '%s' OR "
+                                "description MATCH '%s' OR "
+                                "manufacturer MATCH '%s' OR "
+                                "year LIKE '%s')",
+                                search_text, search_text,
+                                search_text, search_text,
+                                search_text, search_text,
+                                search_text);
+                else
+                        g_string_append (expression, "name ISNULL");
+        }
+
+        g_free (column_name);
+        g_free (search_text);
+}
 
 static gboolean
-tree_view_show_popup_menu (GdkEventButton *event)
+tree_view_show_popup_menu (GdkEventButton *event,
+                           GtkTreeViewColumn *column)
 {
         GtkMenu *menu;
+        GtkWidget *view;
+        const gchar *column_name;
+        const gchar *column_title;
+        GvaGameStoreColumn column_id;
+        gchar *label, *tooltip;
+
+        view = gtk_tree_view_column_get_tree_view (column);
+        g_object_set_data (G_OBJECT (view), "popup-menu-column", column);
+        column_name = g_object_get_data (G_OBJECT (column), "name");
+        gva_columns_lookup_id (column_name, &column_id);
+        column_title = gva_columns_lookup_title (column_id);
+
+        /* Update the "Remove Column" item in the popup menu. */
+        label = g_strdup_printf (
+                _("Remove %s Column"), column_title);
+        tooltip = g_strdup_printf (
+                _("Remove the \"%s\" column from the game list"),
+                column_title);
+        g_object_set (
+                GVA_ACTION_REMOVE_COLUMN, "label",
+                label, "tooltip", tooltip, NULL);
+        g_free (tooltip);
+        g_free (label);
 
         menu = GTK_MENU (gva_ui_get_managed_widget ("/game-popup"));
 
@@ -100,32 +161,23 @@ tree_view_search_equal (GtkTreeModel *model,
                         const gchar *key,
                         GtkTreeIter *iter)
 {
-        const gchar *valid_chars;
-        gchar **str_array;
         gchar *title;
+        gchar *s1, *s2;
         gboolean retval;
 
-        /* Filter out spaces and punctuation so that, e.g. typing
-         * "mspacman" will match "Ms. Pac-Man". */
-
-        valid_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-                      "abcdefghijklmnopqrstuvwxyz"
-                      "0123456789";
-
         g_assert (column == GVA_GAME_STORE_COLUMN_DESCRIPTION);
         gtk_tree_model_get (model, iter, column, &title, -1);
         g_assert (title != NULL);
 
-        g_strcanon (title, valid_chars, '?');
-        str_array = g_strsplit_set (title, "?", -1);
-        g_free (title);
-        title = g_strjoinv (NULL, str_array);
-        g_strfreev (str_array);
+        s1 = gva_search_collate_key (key);
+        s2 = gva_search_collate_key (title);
 
         /* Return FALSE if the row matches. */
-        retval = (g_ascii_strncasecmp (key, title, strlen (key)) != 0);
+        retval = (strncmp (s1, s2, strlen (s1)) != 0);
 
         g_free (title);
+        g_free (s1);
+        g_free (s2);
 
         return retval;
 }
@@ -142,7 +194,6 @@ gva_tree_view_init (void)
 {
         GtkTreeView *view;
         GtkMenu *menu;
-        gchar *name;
 
         view = GTK_TREE_VIEW (GVA_WIDGET_MAIN_TREE_VIEW);
 
@@ -154,6 +205,12 @@ gva_tree_view_init (void)
                 G_CALLBACK (tree_view_selection_changed_cb), NULL);
 
         gva_columns_load (view);
+        gva_ui_add_column_actions (view);
+        gva_tree_view_update_status_bar ();
+
+        gtk_tree_view_set_search_equal_func (
+                view, (GtkTreeViewSearchEqualFunc)
+                tree_view_search_equal, NULL, NULL);
 }
 
 /**
@@ -190,28 +247,40 @@ gva_tree_view_lookup (const gchar *game)
 gboolean
 gva_tree_view_update (GError **error)
 {
+        GString *expression;
         const gchar *name;
-        const gchar *expr;
+        gboolean success;
+
+        expression = g_string_sized_new (128);
 
         switch (gva_tree_view_get_selected_view ())
         {
                 case 0:  /* Available Games */
-                        expr = NULL;
                         break;
 
                 case 1:  /* Favorite Games */
-                        expr = "isfavorite(name) == \"yes\"";
+                        g_string_append (expression, "favorite == \"yes\"");
                         break;
 
                 case 2:  /* Search Results */
-                        expr = "name == NULL";
+                        tree_view_add_search_expression (expression);
                         break;
 
                 default:
                         g_assert_not_reached ();
         }
 
-        if (!gva_tree_view_run_query (expr, error))
+        if (!gva_preferences_get_show_clones ())
+        {
+                if (expression->len > 0)
+                        g_string_append (expression, " AND ");
+                g_string_append (expression, "cloneof ISNULL");
+        }
+
+        success = gva_tree_view_run_query (expression->str, error);
+        g_string_free (expression, TRUE);
+
+        if (!success)
                 return FALSE;
 
         name = gva_tree_view_get_last_selected_game ();
@@ -223,7 +292,7 @@ gva_tree_view_update (GError **error)
 
 /**
  * gva_tree_view_run_query:
- * @expr: an SQL "where" expression
+ * @expression: an SQL "where" expression
  * @error: return location for a #GError, or %NULL
  *
  * Similar to gva_tree_view_update() but applies custom criteria to the game
@@ -232,16 +301,13 @@ gva_tree_view_update (GError **error)
  * Returns: %TRUE on success, %FALSE if an error occurred
  **/
 gboolean
-gva_tree_view_run_query (const gchar *expr,
+gva_tree_view_run_query (const gchar *expression,
                          GError **error)
 {
         GvaGameStoreColumn column_id;
         GtkSortType order;
         GtkTreeView *view;
         GtkTreeModel *model;
-        GdkCursor *cursor;
-        GdkDisplay *display;
-        GdkWindow *window;
         GString *string;
         GSList *list;
         const gchar **strv;
@@ -252,14 +318,11 @@ gva_tree_view_run_query (const gchar *expr,
         view = GTK_TREE_VIEW (GVA_WIDGET_MAIN_TREE_VIEW);
 
         /* Build a comma-separated list of column names. */
-        list = gva_columns_get_names (view, FALSE);
+        list = gva_columns_get_names_full (view);
         strv = g_new0 (const gchar *, g_slist_length (list) + 1);
         while (list != NULL)
         {
-                if (strcmp (list->data, "favorite") == 0)
-                        strv[ii++] = "isfavorite(name) AS favorite";
-                else
-                        strv[ii++] = list->data;
+                strv[ii++] = list->data;
                 list = g_slist_delete_link (list, list);
         }
         columns = g_strjoinv (", ", (gchar **) strv);
@@ -269,21 +332,22 @@ gva_tree_view_run_query (const gchar *expr,
         g_string_printf (string, SQL_SELECT_GAMES, columns);
         g_free (columns);
 
-        if (expr != NULL)
-                g_string_append_printf (string, " AND (%s)", expr);
+        if (expression != NULL && *expression != '\0')
+                g_string_append_printf (string, " WHERE %s", expression);
 
-        window = gtk_widget_get_parent_window (GTK_WIDGET (view));
-        display = gtk_widget_get_display (GTK_WIDGET (view));
-        cursor = gdk_cursor_new_for_display (display, GDK_WATCH);
-
-        gdk_window_set_cursor (window, cursor);
         gtk_widget_set_sensitive (GTK_WIDGET (view), FALSE);
+        gva_main_cursor_busy ();
+
         model = gva_game_store_new_from_query (string->str, error);
-        sensitive = (gtk_tree_model_iter_n_children (model, NULL) > 0);
-        gtk_widget_set_sensitive (GTK_WIDGET (view), sensitive);
-        gdk_window_set_cursor (window, NULL);
+
+        /* Don't touch widgets if gtk_main_quit() has been called. */
+        if (!gtk_main_iteration_do (FALSE))
+        {
+                sensitive = (gtk_tree_model_iter_n_children (model, NULL) > 0);
+                gtk_widget_set_sensitive (GTK_WIDGET (view), sensitive);
+                gva_main_cursor_normal ();
+        }
 
-        gdk_cursor_unref (cursor);
         g_string_free (string, TRUE);
 
         if (model == NULL)
@@ -300,8 +364,15 @@ gva_tree_view_run_query (const gchar *expr,
 
         gtk_tree_view_set_model (view, model);
         gtk_tree_view_columns_autosize (view);
+        gva_tree_view_update_status_bar ();
         g_object_unref (model);
 
+        /* Need to reset the search column after loading a new model,
+         * since GtkTreeView apparenly forgets.  This is not mentioned
+         * in the GTK+ documentation (GNOME bug #351910). */
+        gtk_tree_view_set_search_column (
+                view, GVA_GAME_STORE_COLUMN_DESCRIPTION);
+
         return TRUE;
 }
 
@@ -325,6 +396,83 @@ gva_tree_view_get_model (void)
         return gtk_tree_view_get_model (view);
 }
 
+/**
+ * gva_tree_view_update_status_bar:
+ *
+ * Puts a message in the main status bar containing the MAME version and the
+ * number of games displayed in the current view.  This message gets shown
+ * when the status bar has nothing more important to show.
+ **/
+void
+gva_tree_view_update_status_bar (void)
+{
+        GtkTreeView *view;
+        GtkTreeModel *model;
+        GString *message;
+        const gchar *units;
+        gchar *mame_version;
+        guint context_id;
+        gint count = 0;
+        GError *error = NULL;
+
+        view = GTK_TREE_VIEW (GVA_WIDGET_MAIN_TREE_VIEW);
+        model = gtk_tree_view_get_model (view);
+        message = g_string_sized_new (128);
+
+        mame_version = gva_mame_get_version (&error);
+        gva_error_handle (&error);
+
+        if (model != NULL)
+                count = gtk_tree_model_iter_n_children (model, NULL);
+
+        switch (gva_tree_view_get_selected_view ())
+        {
+                case 0:
+                        units = ngettext (
+                                "Available Game",
+                                "Available Games",
+                                count);
+                        break;
+
+                case 1:
+                        units = ngettext (
+                                "Favorite Game",
+                                "Favorite Games",
+                                count);
+                        break;
+
+                case 2:
+                        units = ngettext (
+                                "Search Result",
+                                "Search Results",
+                                count);
+                        break;
+
+                default:
+                        g_return_if_reached ();
+        }
+
+        if (mame_version != NULL)
+        {
+                g_string_append (message, mame_version);
+                g_free (mame_version);
+
+                if (count > 0)
+                {
+                        /* Append a UTF-8 encoded bullet. */
+                        g_string_append (message, "  •  ");
+                }
+        }
+
+        if (count > 0)
+                g_string_append_printf (message, "%d %s", count, units);
+
+        context_id = gva_main_statusbar_get_context_id (G_STRFUNC);
+        gva_main_statusbar_pop (context_id);
+        gva_main_statusbar_push (context_id, message->str);
+        g_string_free (message, TRUE);
+}
+
 /**
  * gva_tree_view_get_selected_game:
  *
@@ -443,8 +591,8 @@ gva_tree_view_set_selected_view (gint view)
  * gva_tree_view_get_last_selected_game:
  *
  * Returns the name of the most recently selected game in either the
- * current or the previous session of GNOME Video Arcade.  Note that
- * it is still possible for this function to return %NULL.
+ * current or the previous session of <emphasis>GNOME Video Arcade</emphasis>.
+ * Note that it is still possible for this function to return %NULL.
  *
  * Returns: the name of the most recently selected game
  **/
@@ -475,12 +623,12 @@ gva_tree_view_get_last_selected_game (void)
  * gva_tree_view_set_last_selected_game:
  * @game: the name of a game
  *
- * Writes %game to GConf key
+ * Writes @game to GConf key
  * <filename>/apps/gnome-video-arcade/selected-game</filename>.
  *
  * This is used to remember which game was selected in the previous
- * session of GNOME Video Arcade so that the same game can be selected
- * again automatically at startup.
+ * session of <emphasis>GNOME Video Arcade</emphasis> so that the same
+ * game can be selected again automatically at startup.
  **/
 void
 gva_tree_view_set_last_selected_game (const gchar *game)
@@ -503,7 +651,8 @@ gva_tree_view_set_last_selected_game (const gchar *game)
  * @order: return location for the sort order
  *
  * Retrieves the most recently selected sort column ID and sort order in
- * either the current or the previous session of GNOME Video Arcade.
+ * either the current or the previous session of
+ * <emphasis>GNOME Video Arcade</emphasis>.
  **/
 void
 gva_tree_view_get_last_sort_column_id (GvaGameStoreColumn *column_id,
@@ -540,8 +689,8 @@ gva_tree_view_get_last_sort_column_id (GvaGameStoreColumn *column_id,
  * <filename>/apps/gnome-video-arcade/sort-column</filename>.
  *
  * This is used to remember which how the game list view was sorted in
- * the previous session of GNOME Video Arcade so that the same sorting
- * can be applied again automatically at startup.
+ * the previous session of <emphasis>GNOME Video Arcade</emphasis> so that
+ * the same sorting can be applied again automatically at startup.
  **/
 void
 gva_tree_view_set_last_sort_column_id (GvaGameStoreColumn column_id,
@@ -582,19 +731,21 @@ gva_tree_view_button_press_event_cb (GtkTreeView *view,
 {
         if (event->button == 3 && event->type == GDK_BUTTON_PRESS)
         {
+                GtkTreeViewColumn *column;
                 GtkTreePath *path;
                 gboolean valid;
 
                 /* Select the row that was clicked. */
                 valid = gtk_tree_view_get_path_at_pos (
-                        view, event->x, event->y, &path, NULL, NULL, NULL);
+                        view, event->x, event->y, &path,
+                        &column, NULL, NULL);
                 if (valid)
                 {
                         gtk_tree_view_set_cursor (view, path, NULL, FALSE);
                         gtk_widget_grab_focus (GTK_WIDGET (view));
                         gtk_tree_path_free (path);
 
-                        return tree_view_show_popup_menu (event);
+                        return tree_view_show_popup_menu (event, column);
                 }
         }
 
@@ -614,7 +765,54 @@ gva_tree_view_button_press_event_cb (GtkTreeView *view,
 gboolean
 gva_tree_view_popup_menu_cb (GtkTreeView *view)
 {
-        return tree_view_show_popup_menu (NULL);
+        GtkTreeViewColumn *column;
+
+        gtk_tree_view_get_cursor (view, NULL, &column);
+
+        return tree_view_show_popup_menu (NULL, column);
+}
+
+/**
+ * gva_tree_view_query_tooltip_cb:
+ * @view: the main tree view
+ * @x: the x coordinate of the cursor position where the request has
+ *     been emitted
+ * @y: the y coordinate of the cursor position where the request has
+ *     been emitted
+ * @keyboard_mode: %TRUE if the tooltip was triggered using the keyboard
+ * @tooltip: a #GtkTooltip
+ *
+ * Handles tooltips for the main tree view.
+ *
+ * Returns: %TRUE if @tooltip should be shown right now, %FALSE otherwise
+ **/
+gboolean
+gva_tree_view_query_tooltip_cb (GtkTreeView *view,
+                                gint x,
+                                gint y,
+                                gboolean keyboard_mode,
+                                GtkTooltip *tooltip)
+{
+        GtkTreeViewColumn *column;
+        GtkTreePath *path;
+        GtkTreeIter iter;
+
+        if (!gtk_tree_view_get_tooltip_context (
+                view, &x, &y, keyboard_mode, NULL, &path, &iter))
+                return FALSE;
+
+        /* Figure out which column we're pointing at. */
+        if (keyboard_mode)
+                gtk_tree_view_get_cursor (view, NULL, &column);
+        else
+                gtk_tree_view_get_path_at_pos (
+                        view, x, y, NULL, &column, NULL, NULL);
+
+        /* Restrict the tip area to a single cell. */
+        gtk_tree_view_set_tooltip_cell (view, tooltip, path, column, NULL);
+
+        return gva_columns_query_tooltip (column, path, tooltip);
+
 }
 
 /**
